import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter_query/src/types.dart';
import 'package:rxdart/subjects.dart';

class Query<T> {
  final BehaviorSubject<QueryState<T>> stateStream =
      BehaviorSubject.seeded(QueryState<T>(), sync: true);

  QueryConfig config;

  Query(this.config);

  Future<T> fetch() {
    Future<T> queryFn;
    if (!stateStream.value.hasData) {
      stateStream.add(QueryState(
        snapshot: AsyncSnapshot<T>.waiting(),
      ));
    }

    if (config.isInfinite) {
      final cursor = config.getFetchMore(stateStream.value.data);
      if (cursor != false) {
        queryFn = config.fn(cursor);
      }
    } else {
      queryFn = config.fn();
    }

    return queryFn.then((data) {
      var current = stateStream.value.data;
      var canFetchMore;
      if (config.isInfinite) {
        if (current is List && data is List) {
          current.addAll(data as dynamic);
        }
        canFetchMore = data is List ? data.length > 0 : false;
      }

      stateStream.add(
        QueryState(
          snapshot: AsyncSnapshot<T>.withData(ConnectionState.active, data),
          canFetchMore: canFetchMore,
        ),
      );
      return data;
    }).catchError((err) {
      stateStream.add(
        QueryState(
          snapshot: AsyncSnapshot<T>.withError(ConnectionState.done, err),
        ),
      );
    });
  }

  void dispose() {
    stateStream.close();
  }
}

class QueryState<T> {
  AsyncSnapshot<T> snapshot;
  QueryStatus status;
  int failureCount;
  bool isInvalidated;
  bool canFetchMore;

  bool get isError {
    return snapshot?.hasError;
  }

  bool get isFetching {
    return snapshot?.connectionState == ConnectionState.waiting;
  }

  bool get isIdle {
    // print('isIdle needs implementation');
    return false;
  }

  bool get isLoading {
    return isFetching && !(snapshot?.hasData ?? false);
  }

  bool get isSuccess {
    return snapshot?.connectionState == ConnectionState.active &&
        !isError &&
        (snapshot?.hasData ?? false);
  }
  //  isPreviousData,
  //  isStale,

  T get data {
    return snapshot.data;
  }

  bool get hasData {
    return snapshot.data != null;
  }

  QueryState({
    this.snapshot = const AsyncSnapshot.nothing(),
    this.failureCount = 0,
    this.canFetchMore = false,
  });
}

class QueryConfig {
  final Function fn;
  final Function getFetchMore;
  final bool isInfinite;
  final String queryKey;
  final Function onSuccess;
  final Function onError;
  final Function onSettled;

  QueryConfig({
    @required this.fn,
    this.getFetchMore,
    this.isInfinite = false,
    this.queryKey,
    this.onSuccess,
    this.onError,
    this.onSettled,
  });
}
